/***********************************************************************
 * system-demo.c 
 * 
 * Version: v.0.3.0
 *
 * Description
 *
 * This MicroBlaze program samples the sound from the AC'97 controller,
 * sends it to the FFT hardware, retrieves the FFT data, and determines
 * whether or not a music note has been played or stopped playing.
 * Upon detecting any music note has started or stopped playing, the
 * software will display the corresponding MIDI sequences to a terminal
 * program via the RS232 port.
 *
 * In this version, capturing sound, loading the FFT, unloading the FFT,
 * and determine the peak note is performed simultaneously.
 *
 * A note on FFT
 *
 * The lines that sends and retrieves data to and from the FFT hardware
 * block is commented out in this version because our FFT hardware block
 * is not yet functional. A function called hardcoded_fft() is used in
 * place of the function that retrieves FFT data from the FFT hardware.
 * This function generates hard-coded FFT data in order to test and
 * demonstrate the functionalities of this software.
 *
 * Limitations
 *
 * The current implementation assumes only one note is playing at a time.
 * If two notes are playing at the same time, only the loudest note will
 * be outputted to the MIDI device.
 * 
 * Another limitation is that this software cannot detect any note
 * below the third octave. This is because notes below the third
 * octave may appear in the same slice of the FFT data, making it
 * impossible to distinguish the note.
 *
 ***********************************************************************/

#include "xparameters.h"
#include "xbasic_types.h" //Xilinx basic types for device drivers
#include "xac97.h"
#include "xac97_l.h"
#include "xac97_i.h"
#include "mb_interface.h"

#define TRUE 1
#define FALSE 0
#define printf xil_printf
#define THRESHOLD 128
#define printf xil_printf   /* a smaller footprint printf */

/* getNote()
 * Input:   freq_index - index of the FFT output
 * Returns: The note number used in the MIDI standard corresponding to
 *          the given index value.
 */
int getNote(int freq_index)
{
	float freq;
	int octave;
	int note;

	// Return 0 if unsupported note is played
	if( freq_index < 17 )
		return 0;

	freq = freq_index * 7.79724121;

	// Compute the base frequency and octave
	for(octave=3; freq>254.285; octave++)
		freq = freq / 2;

	// Compute the number representing the note, where
	// 0=A, 1=A#, 2=B, 3=C, 4=C#, 5=D,
	// 6=D#, 7=E, 8=F, 9=F#, 10=G, 11=G#
	for ( note=0; freq > 134.7; note++)
		freq = freq / 1.059280855199;

	// Return the MIDI encoding of the note
	return (octave+1) * 12 + note;
}

/* This function returns the velocity of the note based
 * on the given FFT magnitude.
 * The velocity is based on a logarithmic scale.
 */
int getVelocity(int max_val)
{
	if(max_val == 0)
		return 0;
	return getVelocity(max_val/5) + 5;
}

/* This function is used for testing purpose. It returns
 * hard-coded FFT values based on the count and count of
 * the main program loop.
 */
Xuint32 hardcoded_fft(int count, int loop_num)
{
	//int loop_num;
	//loop_num = loop_idx % 100;

	if(0 <= loop_num && loop_num < 20) // C5
	{
		if(count==67)
			return 1024 - loop_num;
		else if (count == 34)
			return 512 - loop_num;
		else if (count == 74)
			return 256 - loop_num;
	}
	else if(20 <= loop_num && loop_num < 40) // G4
	{
		if(count==50)
			return 34359738367 - loop_num;
		else if (count == 25)
			return 17179869183 - loop_num;
		else if (count == 55)
			return 8589934592;
		else if(count == 67)
			return 1024 - loop_num;
		else if (count == 34)
			return 512 - loop_num;
		else if (count == 74)
			return 256 - loop_num;
	}
	else if(40 <= loop_num && loop_num < 60) // G4 played again, C5 released
	{
		if(count==50)
			return 34359738367 - loop_num + 100;
		else if (count == 25)
			return 17179869183 - loop_num;
		else if (count == 55)
			return 8589934592 - loop_num;
	}
	else if(60 <= loop_num && loop_num < 80) // G#4 played, G4 not released
	{
		if(count==54)
			return 2048-loop_num;
		else if (count == 27)
			return 1024;
		else if (count == 59)
			return 512;
		else if(count==50)
			return 34359738367 - loop_num + 100;
		else if (count == 25)
			return 17179869183 - loop_num;
		else if (count == 55)
			return 8589934592 - loop_num;
	}
	else if(80 <= loop_num && loop_num < 100) // No note is playing
	{
		if(count==58)
			return 20;
		else if (count == 267)
			return 35;
		else if (count == 23)
			return 77;
	}
	return 0;
}

int main (void) 
{
	XStatus Status;
	Xuint16 sample;
	Xuint16 count;
	Xuint32 fft_data;
	Xuint32 max_val; // Maximum value of FFT data
	Xuint32 max_val_previous;
	int max_index; // Index of maximum value of FFT data
	int max_index_previous;
	int main_loop_count;
	int state;
	int midi_code;
	int freq;
	int thresh;
	int range;

	state = 0;
	max_val_previous = 0;

	init_sound(8000); // Initialize AC97 Sound controller

	printf("\n\r\n");
	printf("======================================\n\r");
	printf("      Audio to MIDI Encoder Demo\n\r");
	printf("             Version 0.3\n\r");
	printf("======================================\n\r\n");
	printf("  Cycle   MsgType   Note#   Velocity\n\r");
	printf("--------------------------------------\n\r");

	playback_enable();// playback_disable();
	record_enable(); // record_disable();
	for(count=0; count<1024; count++)
	{
		// Read sound data from FSL (AC'97 Controller)
		microblaze_bread_datafsl(sample,0);

		// Write sound data to FFT
		// microblaze_bwrite_datafsl(sample,1);
	}

	for(main_loop_count=0; main_loop_count < 100 ; main_loop_count++)
	{
		printf("\r   %d ",main_loop_count);
		if(main_loop_count<10)
			printf("  ");
		else if(main_loop_count<100)
			printf(" ");

		max_val = 0;
		max_index = 0;

		record_enable();
		for(count=0; count<1024; count++)
		{
			// Read sound data from FSL (AC'97 Controller)
			microblaze_bread_datafsl(sample,0);

			// Write sound data to FFT
			// microblaze_bwrite_datafsl(sample,1);

			// Write sound data to AC'97 Controller (For Demonstration Purpose)
			microblaze_bwrite_datafsl(sample,0);
			
			// Read data from FFT
			// microblaze_bread_datafsl(fft_data,1);
			fft_data = hardcoded_fft(count,main_loop_count);

			// Determine whether the current frequency component
			// has the highest magnitude.
			// Disregarding the extremely low frequency components.
			if(max_val < fft_data && count > 16)
			{
				max_val = fft_data;
				max_index = count;
			}
		}
		record_disable();

		// Stop playing the note if no note is played.
		if (max_val < THRESHOLD)
		{
			// stop playing the previous note if applicable
			if(state != 0)
			{
				printf("    Stop      ");
				midi_code = getNote(max_index_previous);
				printf("%d\n\r",midi_code);
			}
			// printf("No note is currently playing\n\r");
			max_index_previous = 0;
			state = 0;
		}
		// Start playing another note if another note is played
		else if (max_index != max_index_previous)
		{
			if(state!=0)
			{
				// stop playing the previous note
				printf("    Stop      ");
				midi_code = getNote(max_index_previous);
				printf("%d\n\r    ",midi_code);
			}

			// start playing the current note
			if(state!=0)
				printf("       Start     ");
			else
				printf("    Start     ");
			midi_code = getNote(max_index);
			printf("%d        %d\n\r",midi_code,max_val/100);

			max_index_previous = max_index;
			state = 1;
		}
		// If the current note is played again, start playing the note again on MIDI.
		else if ( max_val > max_val_previous )
		{
			// start playing the current note (again)
			printf("    Start     ");
			midi_code = getNote(max_index);
			printf("%d        %d\n\r",midi_code,max_val/100);
			
			state = 2;
		}
		max_val_previous = max_val;
	}
	
	while(TRUE)
	{
		printf("\r============END OF PROGRAM============");
	}
	exit(0);
}