/***********************************************************************
 * system.c 
 * 
 * Version: v.1.3.2
 *
 * Description
 *
 * This MicroBlaze program samples the sound from the AC'97 controller,
 * sends it to the FFT hardware, retrieves the FFT data, and determines
 * whether or not a music note has been played or stopped playing.
 * Upon detecting any music note has started or stopped playing, the
 * software will write the MIDI sequences corresponds to these events
 * to the RS232 serial port.
 * 
 * A note on FFT
 *
 * The lines that sends and retrieves data to and from the FFT hardware
 * block is commented out in this version because our FFT hardware block
 * is not yet functional. A function called hardcoded_fft() is used in
 * place of the function that retrieves FFT data from the FFT hardware.
 * This function generates hard-coded FFT data in order to test and
 * demonstrate the functionalities of this software.
 * 
 * Limitations
 *
 * The current implementation assumes only one note is playing at a time.
 * If two notes are playing at the same time, only the loudest note will
 * be outputted to the MIDI device.
 * 
 * Another limitation is that this software cannot detect any note
 * below the third octave. This is because notes below the third
 * octave may appear in the same slice of the FFT data, making it
 * impossible to distinguish the note.
 *
 ***********************************************************************/

#include "xparameters.h"
#include "xbasic_types.h"
#include "xac97.h"
#include "xac97_l.h"
#include "xac97_i.h"
#include "mb_interface.h"

#define TRUE 1
#define FALSE 0

/* If all FFT data is below this threshold value,
 * this software will assume no note is played. */
#define THRESHOLD 128

/* States indicating whether or not a note is playing */
#define NOT_PLAYING_STATE 0
#define PLAYING_STATE 1

/* Constants used for computing MIDI sequences */
#define STATUS_NOTEOFF 128
#define STATUS_NOTEON 144
#define MIDI_CHANNEL 0

/* This function returns the note number used in the MIDI standard to
 * represent the note that corresponds to the input freq_index value.
 * freq_index is the index of the FFT data, which corresponds to a
 * small range of frequency.
 */
int getNote(int freq_index)
{
	float freq;
	int octave;
	int note;

	// Return 0 if unsupported note is played
	if( freq_index < 17 )
		return 0;

	freq = freq_index * 7.79724121;

	// Compute the base frequency and octave
	for(octave=3; freq>254.285; octave++)
		freq = freq / 2;

	// Compute the number representing the note, where
	// 0=A, 1=A#, 2=B, 3=C, 4=C#, 5=D,
	// 6=D#, 7=E, 8=F, 9=F#, 10=G, 11=G#
	for ( note=0; freq > 134.7; note++)
		freq = freq / 1.059280855199;

	// Return the MIDI encoding of the note
	return (octave+1) * 12 + note;
}

/* This function returns the velocity (loudness) of the note based on the given
 * FFT magnitude. The velocity is an approximation on a logarithmic scale.
 */
int getVelocity(int max_val)
{
	if(max_val == 0)
		return 0;
	return getVelocity(max_val/5) + 5;
}

/* This function is created for testing and demonstration purposes. It returns
 * hard-coded FFT values based on the count and count of the main program loop.
 */
Xuint32 hardcoded_fft(int index, int loop_count)
{
	int loop_num;
	
	loop_num = loop_count % 100;

	/* Output FFT result corresponding to C5 playing. */
	if(0 <= loop_num && loop_num < 20)
	{
		if(index == 67)
			return 1024 - loop_num;
		else if (index == 34)
			return 512 - loop_num;
		else if (index == 74)
			return 256 - loop_num;
	}
	/* Output FFT result corresponding to G4 playing.
	 * C5 is not released at this point. */
	else if(20 <= loop_num && loop_num < 40)
	{
		if(index == 50)
			return 32767 - loop_num;
		else if (index == 25)
			return 16384 - loop_num;
		else if (index == 55)
			return 8192;
		else if(index == 67)
			return 1024 - loop_num;
		else if (index == 34)
			return 512 - loop_num;
		else if (index == 74)
			return 256 - loop_num;
	}
	/* Output FFT result corresponding to G4 played again with loud
	 * volume and C5 released */
	else if(40 <= loop_num && loop_num < 60)
	{
		if(index == 50)
			return 32767 - loop_num + 100;
		else if (index == 25)
			return 16384 - loop_num;
		else if (index == 55)
			return 8192 - loop_num;
	}
	/* Output FFT result corresponding to G#4 playing with a low volume.
	 * G4 is not released at this point. */
	else if(60 <= loop_num && loop_num < 80)
	{
		if(index == 54)
			return 2048-loop_num;
		else if (index == 27)
			return 1024;
		else if (index == 59)
			return 512;
		else if(index == 50)
			return 32767 - loop_num + 100;
		else if (index == 25)
			return 16384 - loop_num;
		else if (index == 55)
			return 8192 - loop_num;
	}
	/* Output FFT result corresponding to a note with very low frequency
	 * started playing. All other notes have stopped playing. */
	else if(80 <= loop_num && loop_num < 100)
	{
		if(index == 58)
			return 20;
		else if (index == 267)
			return 35;
		else if (index == 23)
			return 77;
		else if (index == 10)
			return 4000;
	}
	return 0;
}

int main (void) 
{
	Xuint16 sample;           // Variable for holding sound samples.
	Xuint32 fft_data;         // Variable for holding retrieved FFT data.
	Xuint32 max_val;          // Maximum value of FFT data.
	Xuint32 max_val_previous; // Max value of FFT data in the previous loop.
	int max_index;            // Index of maximum value of FFT data.
	int max_index_previous;   // Index of max value of FFT data in previous loop.
	int count;                // Count for various for loops in the program.
	int main_loop_count;      // Count for main program loop.
	int state;                // State indicating whether a note is playing.
	int midi_code;            // The midi code to represent a note.

	/* Initialize Variables */
	state = NOT_PLAYING_STATE;
	max_val_previous = 0;

	/* Initialize AC97 Sound controller */
	init_sound(8000);

	/* Main program loop */
	for(main_loop_count=0; TRUE ; main_loop_count++)
	{
		max_val = 0;
		max_index = 0;

		record_enable();

		/* Capture Audio Data & write to FFT's FSL */
		for(count=0; count<1024; count++)
		{
			/* Read sound data from FSL (AC'97 Controller) */
			microblaze_bread_datafsl(sample,0);

			/* Write sound data to FFT */
			// microblaze_bwrite_datafsl(sample,1);
		}

		/* Get rid of the gibberish data before disabling record */
		for(count=0;count<50;count++)
			microblaze_bread_datafsl(sample,0);

		record_disable();

		/* Read data from FFT and compute the maximum value and the index of the maximum value. */
		for(count=0; count<1024; count++)
		{
			/* Read data from FFT */
			// microblaze_bread_datafsl(fft_data,1);
			fft_data = hardcoded_fft(count,main_loop_count);

			/* Determine whether the current frequency component has the highest magnitude.
			 * Discarding the extremely low frequency components. */
			if(max_val < fft_data && count > 16)
			{
				max_val = fft_data;
				max_index = count;
			}
		}

		/* If all frequencies are below the threshold level. */
		if (max_val < THRESHOLD)
		{
			/* If a current note is playing, send a stop sequence. */
			if(state != NOT_PLAYING_STATE)
			{
				midi_code = getNote(max_index_previous);
				xil_printf("%c%c%c",STATUS_NOTEOFF+MIDI_CHANNEL,midi_code,0);
				state = NOT_PLAYING_STATE;
			}
			max_index_previous = 0;
		}
		/* If a new note started playing. */
		else if (max_index != max_index_previous)
		{
			/* If a note is previously playing, send a stop sequence for the old note. */
			if(state != NOT_PLAYING_STATE)
			{
				midi_code = getNote(max_index_previous);
				xil_printf("%c%c%c",STATUS_NOTEOFF+MIDI_CHANNEL,midi_code,0);
			}
			/* Send a start sequence for the new note. */
			midi_code = getNote(max_index);
			xil_printf("%c%c%c",STATUS_NOTEON+MIDI_CHANNEL,midi_code,getVelocity(max_val));

			max_index_previous = max_index;
			state = PLAYING_STATE;
		}
		/* If the currently playing note is played again. */
		else if ( max_val > max_val_previous )
		{
			/* Send a start sequence for the current playing note. */
			midi_code = getNote(max_index);
			xil_printf("%c%c%c",STATUS_NOTEON+MIDI_CHANNEL,midi_code,getVelocity(max_val));

			state = PLAYING_STATE;
		}
		max_val_previous = max_val;
	}
	exit(0);
}